//P_TeleportMove
local HakuneTeleportMove = function(mo, x, y, z, set)
	local movefunction = set and P_SetOrigin or P_MoveOrigin or P_TeleportMove
	return movefunction(mo, x, y, z)
end

//3DS hud colormap ^_^ futureproofed for 2.2.14
local HakuneColormap = function(v, mo)
	if P_GetSectorLightLevelAt != nil
		local sector = mo.subsector.sector
		local light = P_GetSectorLightLevelAt(sector, mo.x, mo.y, mo.z)
		return v.getSectorColormap(sector, mo.x, mo.y, mo.z, light)
	else
		return nil
	end
end

//Sounds
sfxinfo[freeslot("sfx_hkkick")].caption = "Kick Thok"
sfxinfo[freeslot("sfx_hkdblj")].caption = "Whoosh"
sfxinfo[freeslot("sfx_hkbnce")].caption = "Bounce"
sfxinfo[freeslot("sfx_hklose")].caption = "/"
sfxinfo[freeslot("sfx_hkstmp")].caption = "Stomp"
sfxinfo[freeslot("sfx_hkreb1")].caption = "Wallkick"
sfxinfo[freeslot("sfx_hkreb2")].caption = "Wallkick+"
sfxinfo[freeslot("sfx_hkflor")].caption = "Bounce"

local TAUNTSOUNDS = 3
for i = 1, TAUNTSOUNDS
	sfxinfo[freeslot("sfx_hktnt" + i)].caption = "Taunt"
end

local WAITSOUNDS = 5
for i = 1, WAITSOUNDS
	sfxinfo[freeslot("sfx_hk3ds" + i)].caption = "/"
end

//Constants
local PREPTIME = 5
local KICKSLIDETIME = PREPTIME*2

//Taunt constants and effect
local TAUNTLENGTH = TICRATE/3
local TAUNTFRAME = D
local TAUNTTICS = TAUNTLENGTH/(TAUNTFRAME + 1)
local TAUNTTCSREMAINDER = TAUNTLENGTH % (TAUNTFRAME + 1)
local TAUNTFRAMES = TAUNTFRAME + 1

freeslot("SPR_HAKN")
for i = A + 1, TAUNTFRAMES
	freeslot("S_HAKUNE_EFFECT" + i)
end

for i = A + 1, TAUNTFRAMES
	local nextstate = S_NULL
	local tics = TAUNTTICS
	if i != TAUNTFRAMES
		nextstate = _G["S_HAKUNE_EFFECT" + (i + 1)]
	else
		tics = $ + TAUNTTCSREMAINDER
	end
	
	states[_G["S_HAKUNE_EFFECT" + i]] = {SPR_HAKN,
	FF_FULLBRIGHT | (i - 1), tics, nil, 1, 0, nextstate}
end

//Uh oh! Death.
freeslot("SPR_HDED", "S_HAKUNE_DEAD")
states[S_HAKUNE_DEAD] = {SPR_HDED, FF_ANIMATE, TICRATE*2, nil, C,  4, S_NULL}

//States and actions
local ALTFALLSTATE = freeslot("S_HAKUNE_ALTFALL")
states[ALTFALLSTATE] = {SPR_PLAY, SPR2_MLEL, 4, A_SetHakunePanim, 0, PA_FALL, ALTFALLSTATE}

function A_SetHakunePanim(mo, var1, var2)
	if gamestate ~= GS_LEVEL then return end
	if var2 < 0
		if mo.momz*P_MobjFlip(mo) < 0
			mo.state = ALTFALLSTATE
			mo.player.pflags = $ & ~PF_JUMPDOWN //???? trying
			mo.player.drawangle = mo.angle //remove some of that jank
		end
	elseif mo.player
		mo.player.panim = var2
	end
end

function A_SetHakuneTaunt(mo, var1, var2)
	if mo.player and mo.sprite2
		mo.frame = ($ &~ FF_FRAMEMASK) | P_RandomKey(skins[mo.skin].sprites[mo.sprite2].numframes)
		
		local sprite2 = mo.sprite2
		local frame = mo.frame
		mo.state = S_PLAY_BOUNCE_LANDING
		mo.sprite2 = sprite2
		mo.frame = frame
		mo.tics = TAUNTLENGTH
		
		//Shield explode
		if (mo.player.powers[pw_shield] & SH_NOSTACK) == SH_ARMAGEDDON
			P_BlackOw(mo.player)
			
		//Focus enemy attention
		else
			local found = false
			local maxdist = FixedMul(RING_DIST*3, mo.scale)
			local thokmobj = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
			thokmobj.radius = maxdist
			searchBlockmap("objects", function(refmobj, mobj)
				local state = states[mobj.state]
				if not (mobj.flags & MF_ENEMY) or not (mobj.flags & MF_SPECIAL)
				or not (mobj.flags & MF_SHOOTABLE) or not mobj.health
				or mobj == mo or mobj.target == mo
				or (mobj.target and not mobj.target.player)
				or (not mobj.target and state.action != A_Look
				and state.action != A_SearchForPlayers)
				or not P_CheckSight(mobj, mo)
					return
				end
				mobj.target = mo
				
				if state.action != A_Look then return end
				local var2 = state.var2
				if not var2
					mobj.state = mobj.info.seestate
					A_PlaySeeSound(mobj)
				elseif var2 == 1 // Only go into seestate
					mobj.state = mobj.info.seestate
				elseif var2 == 2 // Only play seesound
					A_PlaySeeSound(mobj)
				end
			end, thokmobj)
			P_RemoveMobj(thokmobj)
		end
	end
	S_StartSound(mo, sfx_hktnt1 + P_RandomKey(TAUNTSOUNDS))
	local effect = P_SpawnMobjFromMobj(mo, 0,0,0, MT_OVERLAY)
	effect.state = S_HAKUNE_EFFECT1
	effect.target = mo
end
	
local KICKSTATE = freeslot("S_HAKUNE_KICK")
local SLIDESTATE = freeslot("S_HAKUNE_SLIDE")
local BOUNCESTATE = freeslot("S_HAKUNE_BOUNCE")
local BUMPSTATE = freeslot("S_HAKUNE_BUMP")
local KICKPREPSTATE = freeslot("S_HAKUNE_KICKPREP")
local DOUBLESTATE = freeslot("S_HAKUNE_DOUBLEJUMP")
local DOUBLEPREPSTATE = freeslot("S_HAKUNE_DOUBLEPREP")
local DOUBLEFALLSTATE = freeslot("S_HAKUNE_DOUBLEFALL")
local TAUNTSTATE = freeslot("S_HAKUNE_TAUNT")
local WAITSTATE = freeslot("S_HAKUNE_WAIT")
local WAITTICS = 4
local SLIPSTATE = freeslot("S_HAKUNE_SLIP")
local KNOCKBACKSTATE = freeslot("S_HAKUNE_KNOCKBACK")

states[KICKSTATE] = {SPR_PLAY, SPR2_GLID, 2, A_SetHakunePanim, 0, PA_DASH, KICKSTATE}
states[SLIDESTATE] = {SPR_PLAY, SPR2_LAND, 2, A_SetHakunePanim, 0, PA_DASH, SLIDESTATE}
states[BOUNCESTATE] = {SPR_PLAY, SPR2_TIRE, 2, A_SetHakunePanim, 0, PA_DASH, BOUNCESTATE}
states[BUMPSTATE] = {SPR_PLAY, SPR2_CLMB, 2, A_SetHakunePanim, 0, PA_JUMP, BUMPSTATE}
states[KICKPREPSTATE] = {SPR_PLAY, SPR2_CLNG, PREPTIME, A_SetHakunePanim, 0, PA_DASH, KICKSTATE}
states[DOUBLESTATE] = {SPR_PLAY, SPR2_FLY_, 2, A_SetHakunePanim, 0, -1, DOUBLESTATE}
states[DOUBLEPREPSTATE] = {SPR_PLAY, SPR2_CLNG, PREPTIME, A_SetHakunePanim, 0, PA_DASH, DOUBLESTATE}
states[DOUBLEFALLSTATE] = {SPR_PLAY, SPR2_BNCE, 2, A_SetHakunePanim, 0, PA_DASH, DOUBLEFALLSTATE}
states[TAUNTSTATE] = {SPR_PLAY, SPR2_TAL0, TAUNTLENGTH, A_SetHakuneTaunt, 0, 0, TAUNTSTATE}
states[WAITSTATE] = {SPR_PLAY, SPR2_TAL1|FF_SPR2ENDSTATE, WAITTICS, A_SetHakunePanim, S_PLAY_WAIT, PA_IDLE, WAITSTATE}
states[SLIPSTATE] = {SPR_PLAY, SPR2_MLEE, 3, A_SetHakunePanim, 0, PA_DASH, SLIPSTATE}
states[KNOCKBACKSTATE] = {SPR_PLAY, SPR2_TWIN, 350, A_SetHakunePanim, 0, PA_PAIN, S_PLAY_FALL}

//effect definitions

//Chaotix speedtrail code (thanks team combi-rings ^^)
freeslot("SPR_HKST", "S_HAKUNE_SPEEDTRAIL")
local fuse = TICRATE/4
A_ChaotixSpeedTrail = function(mo)
	if not mo.fuse then mo.fuse = fuse end
	local scale = ease.outsine(FRACUNIT*mo.fuse/fuse, FRACUNIT)
	mo.spritexscale = scale + FixedMul(FRACUNIT - scale, abs(sin(mo.rollangle)))
	mo.spriteyscale = scale + FixedMul(FRACUNIT - scale, abs(cos(mo.rollangle)))
end
states[S_HAKUNE_SPEEDTRAIL] = {SPR_HKST, A|FF_PAPERSPRITE, 1, A_ChaotixSpeedTrail, 0, 0, S_HAKUNE_SPEEDTRAIL}

local HakuneSpawnSpeedTrail = function(mo, condition, v)
	if condition == 0 or condition == false then return end -- set condition
	if v == nil then v = {} end
	
	-- get momentum
	if v.momx == nil then v.momx = mo.momx end
	if v.momy == nil then v.momy = mo.momy end
	if v.momz == nil then v.momz = mo.momz end
	-- set up the random positioning
	local hdist = (v.radius or mo.radius*2)/mo.scale
	local vdist = (v.height or mo.height)/mo.scale
	vdist = P_RandomRange(0, $)*FRACUNIT
	
	-- get the angle
	local angle = mo.angle
	if v.momx or v.momy
		angle = R_PointToAngle2(0,0, v.momx, v.momy)
	elseif mo.player
		angle = mo.player.drawangle
	end
	
	-- spawn the trail
	local thok = P_SpawnMobjFromMobj(mo,
	P_RandomRange(-hdist, hdist)*FRACUNIT,
	P_RandomRange(-hdist, hdist)*FRACUNIT,
	vdist, MT_THOK)
	
	-- set the angles and state
	thok.angle = angle
	thok.rollangle = R_PointToAngle2(0,0, FixedHypot(v.momx, v.momy), v.momz)
	thok.state = S_HAKUNE_SPEEDTRAIL
	
	-- set scale and color
	if v.scale then thok.scale = v.scale end
	if v.color
		thok.color = v.color
		thok.colorized = true
	end
	
	-- thrust back a little
	local factor = -4*thok.scale
	P_Thrust(thok, angle, FixedMul(abs(cos(thok.rollangle)), factor))
	P_SetObjectMomZ(thok, FixedMul(sin(thok.rollangle), factor))
end

//ring effect for kick/burst
local BURSTFX = freeslot("MT_HAKUNE_BURST")
local BURSTFUSE = 6 // might change my mind later
local KICKFX = freeslot("MT_HAKUNE_KICKFX") //kick legs for. kick
local WALLFX = freeslot("MT_HAKUNE_WALLKICKFX") //explosion effect on wallhit
freeslot(
	"SPR_HKBS",
	"S_HAKUNE_BURSTFX",
	"SPR_HKIK",
	"S_HAKUNE_KICKFX1",
	"S_HAKUNE_KICKFX2",
	"S_HAKUNE_WALLKICKFX"
)
mobjinfo[BURSTFX] = {
	spawnstate = S_HAKUNE_BURSTFX,
	deathstate = S_NULL,
	flags = MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT
}
states[S_HAKUNE_BURSTFX] = {SPR_HKBS, C|FF_FULLBRIGHT, 1, nil, 0, 0, S_HAKUNE_BURSTFX}
mobjinfo[KICKFX] = {
	spawnstate = S_HAKUNE_KICKFX1,
	deathstate = S_NULL,
	flags = MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT,
	dispoffset = 1
}
states[S_HAKUNE_KICKFX1] = {SPR_HKIK, A|FF_ANIMATE, 16, nil, 3, 1, S_HAKUNE_KICKFX2}
states[S_HAKUNE_KICKFX2] = {SPR_HKIK, A|FF_ANIMATE, -1, nil, 3, 2, S_HAKUNE_KICKFX2}
mobjinfo[WALLFX] = {
	spawnstate = S_HAKUNE_WALLKICKFX,
	deathstate = S_NULL,
	flags = MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT
}
states[S_HAKUNE_WALLKICKFX] = {SPR_HKBS, A|FF_ANIMATE|FF_FULLBRIGHT, 4, nil, 1, 2, S_NULL}

//wallkick burst spawning
local function HakuneWallEffect(mo)
	local kickfx = P_SpawnMobj(mo.x, mo.y, mo.z-(mo.scale*P_MobjFlip(mo)), WALLFX)
	kickfx.color = mo.color
	kickfx.scale = mo.scale*2
end

//kick spawning
local function HakuneKickCircle(mo, cont, tics, fade)
	local kick = P_SpawnMobj(mo.x, mo.y, mo.z, KICKFX)
	if cont and cont == S_HAKUNE_KICKFX2
		kick.state = S_HAKUNE_KICKFX2
	elseif cont and tics
		kick.tics = tics
	end
	if fade
		P_KillMobj(kick)
		return
	end
	kick.tracer = mo
	mo.tracer = kick
	kick.color = mo.color
	kick.health = 1
	kick.scale = mo.scale
	kick.flags2 = $|MF2_LINKDRAW
end

//kick think (teleport on player)
local function HakuneKickThink(mo)
	if mo.valid and mo.tracer and mo.tracer.player and mo.tracer.state == KICKSTATE
		P_MoveOrigin(mo, mo.tracer.x, mo.tracer.y, mo.tracer.z)
		mo.health = $+1
		mo.color = mo.tracer.color
		mo.spriteyscale = mo.tracer.spriteyscale
		mo.spritexscale = mo.tracer.spritexscale
		if mo.tracer.eflags & MFE_VERTICALFLIP
			mo.eflags = $ | MFE_VERTICALFLIP
			mo.spriteyoffset = -30*FRACUNIT
		else
			mo.eflags = $ & ~MFE_VERTICALFLIP
			mo.spriteyoffset = 0
		end
		
		if mo.health == 36
			mo.fuse = TICRATE-5
		end
		if mo.health > TICRATE and leveltime % 2 == 0
			mo.frame = $ + FF_TRANS10
		end
	else
		P_KillMobj(mo)
	end
end

addHook("MobjThinker", HakuneKickThink, KICKFX)

//burst circle spawning
local function HakuneBurst(mo, side)
	local sign
	if mo.momz > 0
		sign = 1
	else
		sign = -1
	end
	local burst = P_SpawnMobj(mo.x, mo.y, mo.z+(P_GetPlayerHeight(mo.player)/2), BURSTFX)
	burst.target = mo
	P_SetScale(burst, mo.scale*2)
	burst.destscale = 0
	if sign*P_MobjFlip(mo) < 0
		burst.destscale = mo.scale*8
	end
	burst.fuse = BURSTFUSE
	burst.scalespeed = FRACUNIT/3
	if side == 0 //0 = vertical, 1 = horizontal
		burst.renderflags = $|RF_PAPERSPRITE
		burst.angle = mo.angle+ANGLE_90
		P_InstaThrust(burst, mo.angle+ANGLE_180, 10*FRACUNIT)
	else	
		burst.renderflags = $|RF_FLOORSPRITE
		P_SetObjectMomZ(burst, sign*-10*FRACUNIT)
	end
end
	

//Circle spawning function
local function HakuneCircle(mobj, mobjtype, amount, ns, zoffset, scale, keepmom, color, vertangle)
	zoffset = $ or 0
	local i = 0	while i < amount
		local fa = i*(ANGLE_180/(amount/2))
		local mo = P_SpawnMobjFromMobj(mobj,0,0,zoffset,mobjtype)
		if mo and mo.valid
			local height = mo.height
			mo.scale = scale
			
			if mobjtype == MT_THOK
				mo.state = S_NIGHTSPARKLE1
				mo.tics = 7
			end
			
			if mobj.eflags&MFE_VERTICALFLIP
				mo.z = $ - (mo.height-height)
			end
			
			if color
				mo.color = color
				mo.colorized = true
			end
			
			if vertangle != nil
				mo.momz = FixedMul(sin(fa),ns)/2
				P_InstaThrust(mo, vertangle+ANGLE_90,
				FixedMul(cos(fa),ns)/2)
			else
				mo.momx = FixedMul(sin(fa),ns)/2
				mo.momy = FixedMul(cos(fa),ns)/2
			end
			
			if keepmom
				mo.momx = $ + mobj.momx/keepmom
				mo.momy = $ + mobj.momy/keepmom
				mo.momz = $ + mobj.momz/keepmom
			end
		end
		i = $ + 1
	end
end

//Hakune's slippery floor check (this was painful to figure out but I finally got it)
local function HakuneCheckFloor(mo)
	if not P_IsObjectOnGround(mo) then return false end
	if mo.floorrover and P_MobjFlip(mo) > 0 and mo.floorrover.valid
		local fofline = lines[P_FindSpecialLineFromTag(540, mo.floorrover.master.frontsector.tag)]
		if fofline and fofline.frontside.textureoffset < 0
			return true
		end
	elseif mo.ceilingrover and P_MobjFlip(mo) < 0 and mo.ceilingrover.valid
		local fofline = lines[P_FindSpecialLineFromTag(540, mo.ceilingrover.master.frontsector.tag)]
		if fofline and fofline.frontside.textureoffset < 0
			return true
		end
	else
		local line = lines[P_FindSpecialLineFromTag(540, mo.subsector.sector.tag)]
		if line and line.frontside.textureoffset < 0
			return true
		end
	end
	return false
end

//Can't move conditions
local function CantMove(player)
	local mo = player.mo
	return not (mo and mo.valid and mo.skin == "hakune")
	or P_PlayerInPain(player)
	or player.pflags & PF_SLIDING
	or player.powers[pw_carry]
	or player.spectator
	or player.playerstate
	or not mo.health
	or player.powers[pw_super] == 1
	or player.iseggrobo
	or player.isjettysyn
	or player.guard
	or player.actionstate
	or player.powers[pw_nocontrol]
	or (player.isxmomentum and mo.state == S_PLAY_FACEPLANT)
end

//Reset Hakune
local function ResetHakune(player)
	player.hakuneslide = 0
	player.hakunetauntstate = BOUNCESTATE
	player.hakunekicktimer = 0
	player.hakunetwirlangle = nil
	player.hakunetauntkick = nil
	player.hakunekicktics = nil
	player.hakunekickfade = nil
	player.hakunescreentodraw = $ or 1
	if player.hakune3dstics == nil
		player.hakune3dstics = 0
	else
		player.hakune3dstics = $
	end
	if (player.realmo and not player.mo)
		player.hakunegaming = nil
	end
	
	if player.hakunenoskid
		player.hakunenoskid = false
		if skins[player.skin].name == "hakune"
			player.charflags = $ &~ SF_NOSKID
		end
	end
	
	if player.hakune3ds and not player.powers[pw_carry] == CR_ROLLOUT
		if player.mo
			S_StopSoundByID(player.mo, player.hakune3ds)
		end
		
		S_FadeMusic(100, MUSICRATE, player)
		player.hakune3ds = nil
	end
end
addHook("PlayerSpawn", ResetHakune)
for player in players.iterate ResetHakune(player) end

//Kick junk
local function HakuneKick(mo)
	if mo and mo.valid and mo.skin == "hakune"
	and (not mo.player or mo.player.glidetime < 3)
	and (mo.state == KICKSTATE or mo.state == BOUNCESTATE or mo.state == SLIDESTATE)
	or (mo.state == KICKPREPSTATE and mo.tics == 1)
		return true
	end
end
addHook("PlayerHeight", function(player) if HakuneKick(player.mo) then return P_GetPlayerSpinHeight(player) end end)
addHook("PlayerCanEnterSpinGaps", HakuneKick)
addHook("PlayerCanDamage", function(player)
	local mo = player.mo
	if mo and mo.valid and mo.skin == "hakune" and (HakuneKick(player.mo)
	or (player.hakuneslide == -1 and (mo.state == DOUBLESTATE or mo.state == DOUBLEFALLSTATE)))
		return true
	end
end)

//find if wall hurts IT WORKS NOW!!
local function HakuneInvestigateWall(player, sector)
	local mo = player.mo
	if player.powers[pw_super] or player.powers[pw_flashing] or player.powers[pw_invulnerability]
		return false
	end
	local secspec = GetSecSpecial(sector.special, 1) < 9 and GetSecSpecial(sector.special, 1) > 0
	if not secspec then return false end
	local ph = P_GetPlayerHeight(player)
	for fof in sector.ffloors() do
		--print("there was a fof here")
		local foftop = fof.topheight
		local fofbottom = fof.bottomheight
		if (mo.eflags & MFE_VERTICALFLIP and foftop <= mo.z and fofbottom > mo.z - ph)
		or (mo.eflags & ~MFE_VERTICALFLIP and foftop >= mo.z + ph and fofbottom < mo.z)
			--print("this fof was dangerous and had special " + secspec)
			return true
		end
	end
	return false
end


//Wall kick
local function WallKick(mo, mobj, line)
	local player = HakuneKick(mo) and mo.player
	if not player then return end
	local noclimb = not mobj and line and line.flags & ML_NOCLIMB
	if line
		if line.special == 41 return end
		local side = P_PointOnLineSide(mo.x,mo.y,line)
		local wallsec
		local skysec
		if side == 1
			wallsec = line.frontsector
			skysec = line.backsector
		else
			wallsec = line.backsector
			skysec = line.frontsector
		end
		if wallsec and wallsec.valid
			if HakuneInvestigateWall(player, wallsec) then return end
			if wallsec.ceilingpic:upper():find("F_SKY")
			and wallsec.ceilingheight < (mo.z + mo.height*(min(P_MobjFlip(mo)+1,1)))
				return
			end
		end
	end
	//kick harder
	if mobj and mobj.flags & MF_PUSHABLE
		local hitdir = R_PointToAngle2(0,0, mo.momx, mo.momy)
		local supermult
		if player.powers[pw_super] > 1 then supermult = 3 else supermult = 1 end
		P_Thrust(mobj, hitdir, player.speed/2*supermult)
	end
	HakuneWallEffect(mo)
	mo.state = BUMPSTATE
	if not noclimb and player.glidetime < 2
		player.pflags = $ &~ PF_JUMPED
		P_DoJump(player, false)
		player.pflags = $ &~ PF_STARTJUMP &~ PF_THOKKED &~ PF_NOJUMPDAMAGE
	elseif mo.momz*P_MobjFlip(mo) < 0
		mo.momz = 0
	end
	
	P_BounceMove(mo)
	mo.momx = $/2
	mo.momy = $/2
	
	if player.hakuneslide == -1
		HakuneCircle(mo, MT_IVSP, 8, 8*mo.scale, -32*FRACUNIT, mo.scale/3, 2, nil, player.drawangle)
		player.hakuneslide = 0
		S_StartSound(mo, sfx_hklose)
		
		if noclimb or player.glidetime > 1
			mo.state = BOUNCESTATE
			player.pflags = $ | PF_THOKKED
			player.glidetime = 3	
		else
			S_StartSound(mo, min(sfx_hkreb1 + player.glidetime, sfx_hkreb2))
			player.glidetime = $ + 1
			if player.glidetime >= 3
				mo.state = BOUNCESTATE
				player.pflags = $ | PF_THOKKED
				player.glidetime = max($, 3)
			end
		end
	end
	S_StartSound(mo, sfx_hkbnce)
	
	return //true
end
addHook("MobjMoveBlocked", WallKick, MT_PLAYER)

//Wall Breaker Function
local fofbreak = function(mo, sector, strong)
	if not (mo and sector) then return end
	for fof in sector.ffloors()
		if fof.flags & FF_BUSTUP and fof.flags & FF_EXISTS
		and (strong or not (fof.flags & FF_STRONGBUST))
			local foffloor = fof.topheight
			local fofceiling = fof.bottomheight
			if fof.b_slope and fof.b_slope.valid
				fofceiling = P_GetZAt(fof.b_slope, mo.x, mo.y)
			end
			if fof.t_slope and fof.t_slope.valid
				foffloor = P_GetZAt(fof.t_slope, mo.x, mo.y)
			end
			if foffloor >= mo.z and fofceiling <= mo.z+mo.height
				EV_CrumbleChain(nil, fof)
				if fof.master.flags & ML_EFFECT5
					P_LinedefExecute(P_AproxDistance(fof.master.dx, fof.master.dy)>>FRACBITS, mo, sector)
				end
				return true
			end
		end
	end
end

//Wall Breaker
addHook("MobjLineCollide", function(mo, line)
	if not (line and line.valid and HakuneKick(mo)) return end
	local killone = false
	local killtwo = false
	if line.frontsector and line.frontsector.valid then killone = fofbreak(mo, line.frontsector) end
	if line.backsector and line.backsector.valid then killtwo = fofbreak(mo, line.backsector) end
	if killone or killtwo
		WallKick(mo, nil, line)
	end
end, MT_PLAYER)

//Kick bounce floor invinciblity
/*
addHook("ShouldDamage", function(mo, inf, src)
	if not inf and not src
	and HakuneKick(mo) and mo.player.hakuneslide == -1
	and P_IsObjectOnGround(mo)
		return false
	end
end, MT_PLAYER)
DJJ apparently wanted it gone so Now it is*/

//it has to be this hacky because the thunder shield DOESN'T ACT LIKE A NORMAL SHIELD!!
local function HakuneThunderCheck(player)
	if player.powers[pw_shield] == SH_THUNDERCOIN and player.pflags & PF_THOKKED
	and not (player.pflags & PF_NOJUMPDAMAGE) then
		return true
	else return false end
end

//Start the kick or double jump
addHook("AbilitySpecial", function(player)
	local mo = player.mo
	if not (mo and mo.valid and mo.skin == "hakune") then return end
	if CantMove(player)
	or player.pflags & PF_JUMPSTASIS
	or HakuneThunderCheck(player)
	or player.secondjump
	or mo.state == BOUNCESTATE
	then return true end
	
	//Double jump
	if player.pflags & PF_THOKKED and player.glidetime < 3
		mo.state = DOUBLEPREPSTATE
		player.pflags = $ &~ PF_JUMPED &~ PF_SPINNING &~ PF_THOKKED
		player.secondjump = 1
		player.hakunekicktimer = 0
		
		S_StartSound(mo, sfx_hkdblj)
	
	//Kick
	elseif player.glidetime < 3
		mo.state = KICKPREPSTATE
		player.hakunekicktimer = 0
		player.pflags = $ | PF_THOKKED | PF_JUMPSTASIS & ~PF_SPINNING
		
		S_StartSound(mo, sfx_hkkick)
	end
	
	return true
end)

//Start the double jump downwards
addHook("JumpSpinSpecial", function(player)
	local mo = player.mo
	if CantMove(player)
	or player.pflags & (PF_JUMPSTASIS|PF_SPINDOWN|PF_SHIELDABILITY)
	or player.secondjump
	or HakuneThunderCheck(player)
	or not (player.pflags & PF_THOKKED)
	or player.glidetime > 2
	or mo.state == BOUNCESTATE
	then return end

	mo.state = DOUBLEPREPSTATE
	player.pflags = $ &~ PF_JUMPED &~ PF_SPINNING
	player.secondjump = 2
	
	S_StartSound(mo, sfx_hkstmp)
end)

//still hacking around shields...AGHH
addHook("ShieldSpecial", function(player)
	local mo = player.mo
	if not (mo and mo.valid and mo.skin == "hakune") then return end
	if player.secondjump > 0 then return true end
end)

//Command
local HAKUNE_FLOORS = {
	["DISCO1"] = true,
	["DISCO2"] = true,
	["DISCO3"] = true,
	["DISCO4"] = true,
}
local awesomeness = 10
COM_AddCommand("hakuneawesomeness", function(player, arg)
	local mo = player and player.mo
	if not mo then return end
	arg = tonumber($)
	if arg == nil
		CONS_Printf(player, "Not a number.")
	elseif arg == 0
		player.hakunedisco = false
		awesomeness = 0
		if mo.state == S_TAILSOVERLAY_FLY
			mo.state = S_PLAY_WALK
		end
	elseif tonumber(arg) == player.mo.y and not CantMove(player)
	and P_IsObjectOnGround(player.mo)
		mo.state = S_TAILSOVERLAY_FLY
		player.hakunedisco = true
	else
		if arg > awesomeness
			CONS_Printf(player, "Awesomeness increased!")
		elseif arg < awesomeness
			CONS_Printf(player, "Awesomeness decreased!")
		else
			CONS_Printf(player, "Awesomeness retained!")
		end
		awesomeness = arg
	end
end)

//3DS graphic in first person
hud.add(function(v, player, camera)
	local mo = player.mo
	if not (mo and mo.valid and mo.skin == "hakune")
	and not (displayplayer == player) then return end
	local gameplaying = player.hakunegaming != nil and player.hakunegaming*2+1
	local FINAL3DSPOS = 80*FRACUNIT
	local cropwh = {228*FRACUNIT, 168*FRACUNIT}
	local drawx = 160*FRACUNIT
	local H3DSVFLAGS = V_SNAPTOBOTTOM|V_PERPLAYER
	local hk3ds = player.hakune3dstics
	if gameplaying and hk3ds != nil
		local screens = {v.cachePatch("HAKUNE3D" + gameplaying), v.cachePatch("HAKUNE3D" + (gameplaying+1))}
		local screendraw = screens[player.hakunescreentodraw]
		if camera.chase == false
			if hk3ds < 23 and hk3ds > -1
				local pos = ease.outcubic(hk3ds*FRACUNIT/23, 200*FRACUNIT, FINAL3DSPOS)
				v.drawCropped(drawx, pos, FRACUNIT, FRACUNIT, screendraw, H3DSVFLAGS, HakuneColormap(v, mo), 0, 0, cropwh[1], cropwh[2])
			elseif hk3ds > 0
				v.drawCropped(drawx, FINAL3DSPOS, FRACUNIT, FRACUNIT, screendraw, H3DSVFLAGS, HakuneColormap(v, mo), 0, 0, cropwh[1], cropwh[2])
			elseif hk3ds > -10
				local pos = ease.outexpo((hk3ds*-1)*FRACUNIT/10, FINAL3DSPOS, 200*FRACUNIT)
				v.drawCropped(drawx, pos, FRACUNIT, FRACUNIT, screendraw, H3DSVFLAGS, HakuneColormap(v, mo), 0, 0, cropwh[1], cropwh[2])
			end
		end
	end
end)

//Prevent skidding on frame 1 of slide (sorry about this xp)
addHook("PreThinkFrame", function()
	for player in players.iterate()
		local mo = player.mo
		if not (mo and mo.valid and mo.skin == "hakune") then return end
		if mo.state == BOUNCESTATE and mo.eflags & MFE_JUSTHITFLOOR
			player.powers[pw_nocontrol] = 1
		end
	end
end)

//Altfall in fan animation
addHook("PostThinkFrame", function()
	for player in players.iterate()
		local mo = player.mo
		if not (mo and mo.valid and mo.skin == "hakune") then return end
		if player.powers[pw_carry] == CR_FAN and player.panim == PA_FALL
			mo.state = ALTFALLSTATE
			--print("fan guy")
		end
	end
end)

//Main thinker
addHook("PlayerThink", function(player)
	local mo = player.mo
	if not (mo and mo.valid and mo.skin == "hakune") then ResetHakune(player) return end
	
	local prevmomz = player.hakunemomz or 0
	player.hakunemomz = mo.momz
	
	local prevspeed = player.hakunespeed or 0
	if mo.state != KICKPREPSTATE and mo.state != DOUBLEPREPSTATE
		player.hakunespeed = player.speed
	end
	
	if player.cmd.buttons & BT_CUSTOM1 and player.hakunecustom1 != nil
		player.hakunecustom1 = $ + 1
	else
		player.hakunecustom1 = 0
	end
	
	if player.hakune3dstics == nil //for mid-joiners I think
		player.hakune3dstics = 0
	end
	
	//Slippery animation
	if (player.powers[pw_carry] == CR_ROLLOUT or HakuneCheckFloor(mo))
	and (mo.state == S_PLAY_WALK or mo.state == S_PLAY_RUN) and mo.state != SLIPSTATE
		mo.state = SLIPSTATE
	elseif mo.state == SLIPSTATE and player.speed == 0
		mo.state = S_PLAY_STND
	elseif mo.state == SLIPSTATE and player.speed >= FixedMul(player.runspeed, mo.scale)
		mo.state = SLIPSTATE
	end
	if P_IsObjectOnGround(mo) and mo.state == SLIPSTATE and HakuneCheckFloor(mo) == false
		if player.speed >= FixedMul(player.runspeed, mo.scale)
			mo.state = S_PLAY_RUN
		else
			mo.state = S_PLAY_WALK
		end
	end
	
	//Heavy knockback
	if P_PlayerInPain(player) and player.speed > FixedMul(16*FRACUNIT, mo.scale) and mo.state != KNOCKBACKSTATE
		mo.state = KNOCKBACKSTATE
		player.powers[pw_flashing] = 3*TICRATE
	end
	if mo.state == KNOCKBACKSTATE
		player.powers[pw_nocontrol] = 1
	end
	if player.powers[pw_flashing] == 3*TICRATE and player.panim != PA_PAIN
		player.powers[pw_flashing] = 3*TICRATE-1
	end //this technically has vanilla behavior overlap but it's the best check
		//I can do
	
	//Drown ghost
	if mo.state == S_PLAY_DRWN
		if not mo.hakunedead
			mo.hakunedead = true
			
			local ghost = P_SpawnGhostMobj(mo)
			ghost.state = S_HAKUNE_DEAD
			ghost.fuse = ghost.tics
			ghost.momz = P_MobjFlip(mo)*mo.scale*2
			if player.playerstate
				mo.momz = ghost.momz
			end
			
			player.hakuneghost = ghost
			player.hakuneghosttrans = NUMTRANSMAPS*2 - 1
		end
	elseif mo.hakunedead
		mo.hakunedead = false
	end
	
	//Drown ghost transparency
	if player.hakuneghost and player.hakuneghost.valid
	and player.hakuneghosttrans
		player.hakuneghosttrans = $ - 1
		player.hakuneghost.frame = ($ &~ FF_TRANSMASK) | ((player.hakuneghosttrans/2) << FF_TRANSSHIFT)
	end
	
	//3DS hud stuff
	if player.hakune3dstics != nil and player.hakune3dstics % 4 == 0
		if player.hakunescreentodraw == 1
			player.hakunescreentodraw = 2
		else
			player.hakunescreentodraw = 1
		end
	end
	if player.hakune3ds and player.hakune3dstics != nil
		player.hakune3dstics = max($+1,0)
	elseif player.hakune3dstics != nil and player.hakune3dstics > -10
		player.hakune3dstics = min($-1,-1)
	end
	
	
	//Wait Transition
	if mo.state == S_PLAY_STND and mo.tics == 1
		mo.state = WAITSTATE
	end
	
	//3DS playing
	if mo.state == S_PLAY_WAIT
		mo.anim_duration = min($, WAITTICS)
		if not player.hakune3ds
			player.hakunegaming = P_RandomKey(WAITSOUNDS)
			player.hakune3ds = sfx_hk3ds1 + player.hakunegaming
			S_FadeMusic(25, MUSICRATE, player)
		elseif not S_SoundPlaying(mo, player.hakune3ds)
			S_StartSound(mo, player.hakune3ds)
		end
	elseif player.hakune3ds
		S_StopSoundByID(mo, player.hakune3ds)
		S_FadeMusic(100, MUSICRATE, player)
		player.hakune3ds = nil
	end

	//Prevent the rest of the code
	if CantMove(player) then ResetHakune(player) return end
	
	//Taunt
	if mo.state == S_PLAY_BOUNCE_LANDING
		if mo.tics == 1
			mo.state = player.hakunetauntstate or BOUNCESTATE
			if mo.state == KICKSTATE
				HakuneKickCircle(mo, player.hakunetauntkick, player.hakunekicktics, player.hakunekickfade)
			end
				
		else
			player.pflags = $ | PF_FULLSTASIS
		end
		return
	else
		player.hakunetauntstate = player.panim == PA_IDLE and S_PLAY_STND or mo.state
		
		if player.hakunecustom1 == 1 and not (player.pflags & PF_SPINNING)
			if mo.state == KICKSTATE and mo.tracer
				player.hakunetauntkick = mo.tracer.state
				player.hakunekicktics = mo.tracer.tics
				player.hakunekickfade = mo.tracer.frame & FF_TRANSMASK
			end
			mo.state = TAUNTSTATE
			return
		end
	end
	
	//fallback for obscure bugs
	if player.pflags & PF_SPINNING and player.panim != PA_ROLL and player.panim != PA_JUMP
	and player.mo.state != S_PLAY_SPINDASH
		mo.state = S_PLAY_ROLL
		S_StartSound(mo, sfx_spin)
	end
	
	//Dance
	if mo.state == S_TAILSOVERLAY_FLY
		if mo.tics == states[S_TAILSOVERLAY_FLY].tics
			if player.speed > 12*mo.scale
				mo.tics = 2
			else
				mo.tics = 3
			end
		end
		local sector = mo.subsector and mo.subsector.sector
		if ((mo.eflags & MFE_VERTICALFLIP and not HAKUNE_FLOORS[sector.ceilingpic])
			or (not (mo.eflags & MFE_VERTICALFLIP) and not HAKUNE_FLOORS[sector.floorpic]))
			COM_BufInsertText(consoleplayer, "hakuneawesomeness 0")
		end
	elseif P_IsObjectOnGround(mo) and (mo.state == S_PLAY_STND
	or mo.state == WAITSTATE or mo.state == S_PLAY_EDGE
	or mo.state == S_PLAY_WAIT or mo.state == S_PLAY_WALK
	or mo.state == S_PLAY_RUN)
		if player.hakunedisco
			mo.state = S_TAILSOVERLAY_FLY
			if S_MusicName(player) != "disco"
				COM_BufInsertText(consoleplayer, "hakuneawesomeness 0")
			end
		else
			local sector = mo.subsector and mo.subsector.sector
			if sector
			and ((mo.eflags & MFE_VERTICALFLIP and HAKUNE_FLOORS[sector.ceilingpic])
			or (not (mo.eflags & MFE_VERTICALFLIP) and HAKUNE_FLOORS[sector.floorpic]))
				mo.state = S_TAILSOVERLAY_FLY
			elseif S_MusicName(player) == "disco"
				COM_BufInsertText(consoleplayer, "hakuneawesomeness "..mo.y)
			end
		end
	end
	
	//Sparkles
	if player.hakuneslide == -1 and (HakuneKick(mo)
	or mo.state == DOUBLESTATE or mo.state == DOUBLEFALLSTATE)
		local spark = P_SpawnMobjFromMobj(mo, 0,0,0, MT_THOK)
		spark.state = S_NIGHTSPARKLE1
		spark.tics = 7
		
		HakuneTeleportMove(spark, spark.x + mo.momx, spark.y + mo.momy,
		spark.z + mo.momz)
		
		spark.spritexoffset = P_SignedRandom()*FRACUNIT/6
		spark.spriteyoffset = P_SignedRandom()*FRACUNIT/4 + player.height/2
		
		if HakuneKick(mo)
			spark.spriteyoffset = $*2/3
		end
		
		player.hakunenoskid = true
		player.charflags = $ | SF_NOSKID
	elseif player.hakunenoskid
		player.hakunenoskid = false
		player.charflags = $ &~ SF_NOSKID
	end
	
	//Prepping the kick...
	if mo.state == KICKPREPSTATE
		player.pflags = $ | PF_NOJUMPDAMAGE
		mo.momz = -P_GetMobjGravity(mo)
		
		P_InstaThrust(mo, mo.angle, -2*mo.scale)
		player.drawangle = mo.angle
		
		//Start it!
		if mo.tics == 1
			local superspeed = 0
			if player.powers[pw_super] > 0 then superspeed = 5*FRACUNIT end
			prevspeed = max($, FixedMul(player.actionspd+(5*FRACUNIT*superspeed), mo.scale))
			P_InstaThrust(mo, mo.angle, prevspeed)
			HakuneCircle(mo, MT_IVSP, 8, 8*mo.scale, -32*FRACUNIT, mo.scale/3, 2, nil, mo.angle)
			HakuneBurst(mo, 0)
			HakuneKickCircle(mo)
			player.hakuneslide = -1
			player.hakunekicktimer = 16
		else
			player.pflags = $ | PF_JUMPSTASIS
		end
		
	//Mid kick
	elseif HakuneKick(mo)
		mo.friction = FRACUNIT
		
		//Ground slide
		if P_IsObjectOnGround(mo)
			if player.hakuneslide < 0
				mo.momz = -prevmomz
				mo.z = $ + P_MobjFlip(mo)
				player.pflags = $ | PF_JUMPED | PF_STARTJUMP | PF_THOKKED
				
				HakuneCircle(mo, MT_DUST, 8, 8*mo.scale, -32*FRACUNIT, mo.scale, 2)
				S_StartSound(mo, sfx_hkflor)
				player.hakunekicktimer = 0
				mo.state = BOUNCESTATE
			else
				if mo.state != SLIDESTATE
					mo.state = SLIDESTATE
				end
				
				if player.hakuneslide > KICKSLIDETIME
				or not player.speed
					mo.state = S_PLAY_WALK
					player.hakuneslide = 0
				else
					P_SpawnSkidDust(player, mo.radius, true)
					player.pflags = $ | PF_STASIS
				end
			end
			
			player.hakuneslide = $ + 1
			
		//Air time
		else
			player.pflags = $ | PF_JUMPED | PF_THOKKED
			if mo.state != BOUNCESTATE and mo.state != SLIDESTATE
				mo.momz = $ - P_GetMobjGravity(mo)/2
			end //is this smart
			if player.hakuneslide > 0
				player.hakuneslide = 0
			end
			if player.hakunekicktimer > 0
				player.hakunekicktimer = $ - 1
				if player.hakunekicktimer > 8
					HakuneSpawnSpeedTrail(mo, (player.speed > FixedMul(8*FRACUNIT, mo.scale) or mo.momz > FixedMul(8*FRACUNIT, mo.scale)))
				end
				mo.movefactor = FRACUNIT/5
				mo.state = KICKSTATE //speed it up a lil
			end
		end
		
	//Prepping the double jump...
	elseif mo.state == DOUBLEPREPSTATE
		mo.momx = 0
		mo.momy = 0
		mo.momz = P_GetMobjGravity(mo)*4
		if player.secondjump == 2
			mo.momz = $*-2
		end
		
		//Start it!
		if mo.tics == 1
			P_DoJump(player, false)
			player.hakunekicktimer = 16
			mo.momz = $*3/2
			player.hakunetwirlangle = player.drawangle
			if player.secondjump == 2
				mo.momz = $*-2
				mo.state = DOUBLEFALLSTATE
				prevspeed = $*2/3
			else
				prevspeed = $/2
			end
			
			P_InstaThrust(mo, mo.angle, prevspeed)
			
			HakuneCircle(mo, MT_IVSP, 8, 8*mo.scale, -32*FRACUNIT, mo.scale/3, 2)
			HakuneBurst(mo, 1)
			player.hakuneslide = -1
		else
			player.pflags = $ | PF_JUMPSTASIS | PF_NOJUMPDAMAGE
		end
	
	//double jump twirl
	elseif mo.state == DOUBLESTATE and not P_IsObjectOnGround(mo)
		HakuneSpawnSpeedTrail(mo, (player.speed > FixedMul(8*FRACUNIT, mo.scale) or FixedMul(8*FRACUNIT, mo.scale)))
		if player.hakunekicktimer > 0
			player.hakunekicktimer = $ - 1
			mo.movefactor = FRACUNIT/2
		end
		player.hakunetwirlangle = $-ANG10-FixedAngle(min(55*FRACUNIT,FixedDiv(mo.momz*P_MobjFlip(mo), mo.scale)*4))
		player.drawangle = player.hakunetwirlangle
	//altfall hiccups
	elseif mo.state == ALTFALLSTATE and P_IsObjectOnGround(mo)
		if player.speed == 0
			mo.state = S_PLAY_STND
		elseif player.speed < FixedMul(player.runspeed, mo.scale)
			mo.state = S_PLAY_WALK
		else
			mo.state = S_PLAY_RUN
		end
	//Mid double fall
	elseif mo.state == DOUBLEFALLSTATE
		HakuneSpawnSpeedTrail(mo)
		mo.movefactor=FRACUNIT/2
		if (P_IsObjectOnGround(mo)
		or mo.momz*P_MobjFlip(mo) > 0 or P_IsObjectInGoop(mo))
			HakuneCircle(mo, MT_IVSP, 8, 4*mo.scale, -32*FRACUNIT, mo.scale/3, 2)
			S_StartSound(mo, sfx_hklose)
			S_StartSound(mo, sfx_hkbnce)
			
			player.hakuneslide = 0
			mo.state = ALTFALLSTATE
			
			if P_IsObjectOnGround(mo)
				//Destroy floors under us
				fofbreak(mo, mo.subsector and mo.subsector.sector, true)
				mo.momz = -P_GetMobjGravity(mo)*8
				mo.z = $ + P_MobjFlip(mo)
				S_StopSoundByID(mo, sfx_hkstmp)
				if player.cmd.buttons & BT_SPIN
					mo.state = S_PLAY_ROLL
					player.pflags = $ | PF_SPINNING | PF_THOKKED
					S_StartSound(mo, sfx_spin)
					return
				end
			end
			
			if not P_IsObjectInGoop(mo)
				player.secondjump = 0
				player.pflags = $ | PF_JUMPED | PF_NOJUMPDAMAGE &~ PF_THOKKED
			end
		end
	end
end)